"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.optionalChangeHandler = exports.optionalFieldIntoDelta = exports.optionalFieldEditor = exports.optionalChangeRebaser = void 0;
const core_utils_1 = require("@fluidframework/core-utils");
const index_js_1 = require("../../util/index.js");
const index_js_2 = require("../modular-schema/index.js");
const deltaUtils_js_1 = require("../deltaUtils.js");
const optionalFieldCodecs_js_1 = require("./optionalFieldCodecs.js");
class RegisterMap {
    constructor() {
        this.nestedMapData = new index_js_1.SizedNestedMap();
    }
    clone() {
        const clone = new RegisterMap();
        for (const [id, t] of this.entries()) {
            clone.set(id, t);
        }
        return clone;
    }
    set(id, childChange) {
        if (id === "self") {
            this.nestedMapData.set("self", undefined, childChange);
        }
        else {
            this.nestedMapData.set(id.localId, id.revision, childChange);
        }
    }
    get(id) {
        return id === "self"
            ? this.nestedMapData.tryGet(id, undefined)
            : this.nestedMapData.tryGet(id.localId, id.revision);
    }
    has(id) {
        return this.get(id) !== undefined;
    }
    delete(id) {
        return id === "self"
            ? this.nestedMapData.delete("self", undefined)
            : this.nestedMapData.delete(id.localId, id.revision);
    }
    keys() {
        const changeIds = [];
        for (const [localId, nestedMap] of this.nestedMapData) {
            if (localId === "self") {
                changeIds.push("self");
            }
            else {
                for (const [revisionTag, _] of nestedMap) {
                    changeIds.push(revisionTag === undefined
                        ? { localId }
                        : { localId, revision: revisionTag });
                }
            }
        }
        return changeIds;
    }
    values() {
        return this.nestedMapData.values();
    }
    entries() {
        const entries = [];
        for (const changeId of this.keys()) {
            if (changeId === "self") {
                const entry = this.nestedMapData.tryGet("self", undefined);
                (0, core_utils_1.assert)(entry !== undefined, 0x770 /* Entry should not be undefined when iterating keys. */);
                entries.push(["self", entry]);
            }
            else {
                const entry = this.nestedMapData.tryGet(changeId.localId, changeId.revision);
                (0, core_utils_1.assert)(entry !== undefined, 0x771 /* Entry should not be undefined when iterating keys. */);
                entries.push([changeId, entry]);
            }
        }
        return entries;
    }
    get size() {
        return this.nestedMapData.size;
    }
}
/**
 * Attempt to deduce whether the "self" register of a change is empty or filled in the input context.
 * Optional changesets do not always carry enough information to determine this, in which case this
 * returns `undefined`.
 */
function tryInferInputContext(change) {
    if (change.reservedDetachId !== undefined) {
        return "empty";
    }
    for (const [src] of change.moves) {
        if (src === "self") {
            return "filled";
        }
    }
    return undefined;
}
exports.optionalChangeRebaser = {
    compose: ({ change: change1, revision: revision1 }, { change: change2, revision: revision2 }, composeChild) => {
        const inputContext = tryInferInputContext(change1) ?? tryInferInputContext(change2);
        const earliestReservedDetachId = withRevisionOrUndefined(change1.reservedDetachId, revision1) ??
            withRevisionOrUndefined(change2.reservedDetachId, revision2);
        const { srcToDst, dstToSrc } = getBidirectionalMaps(change1.moves, revision1);
        // This loop will update srcToDst to map from registers before either change to
        // registers after both changes have been applied.
        // dstToSrc maps from registers after change1 is applied to registers before.
        // This loop will not change dstToSrc.
        for (const [src, dst, target] of change2.moves) {
            const srcWithRevision = withRevision(src, revision2);
            const dstWithRevision = withRevision(dst, revision2);
            const originalSrc = dstToSrc.get(srcWithRevision);
            if (originalSrc !== undefined) {
                const entry = srcToDst.get(originalSrc);
                (0, core_utils_1.assert)(entry !== undefined, "There should be a corresponding entry");
                entry[0] = dstWithRevision;
                if (target === "nodeTargeting") {
                    entry[1] = target;
                }
            }
            else {
                srcToDst.set(srcWithRevision, [dstWithRevision, target]);
            }
        }
        const composedMoves = [];
        for (const [src, [dst, target]] of srcToDst.entries()) {
            composedMoves.push([src, dst, target]);
        }
        const childChanges2ByOriginalId = new RegisterMap();
        for (const [id, change] of change2.childChanges) {
            const idWithRevision = withRevision(id, revision2);
            const originalId = dstToSrc.get(idWithRevision);
            childChanges2ByOriginalId.set(originalId ?? idWithRevision, change);
        }
        const composedChildChanges = [];
        for (const [id, childChange1] of change1.childChanges) {
            const idWithRevision = withRevision(id, revision1);
            const childChange2 = childChanges2ByOriginalId.get(idWithRevision);
            composedChildChanges.push([idWithRevision, composeChild(childChange1, childChange2)]);
            childChanges2ByOriginalId.delete(idWithRevision);
        }
        for (const [id, childChange2] of childChanges2ByOriginalId.entries()) {
            composedChildChanges.push([id, composeChild(undefined, childChange2)]);
        }
        const composed = {
            moves: composedMoves,
            childChanges: composedChildChanges,
        };
        if (inputContext === "empty") {
            composed.reservedDetachId = earliestReservedDetachId;
        }
        return composed;
    },
    amendCompose: () => (0, index_js_1.fail)("Not implemented"),
    invert: ({ revision, change }, invertChild, genId) => {
        const { moves, childChanges } = change;
        const invertIdMap = new RegisterMap();
        const withIntention = (id) => {
            if (id === "self") {
                return id;
            }
            return { revision: id.revision ?? revision, localId: id.localId };
        };
        for (const [src, dst] of moves) {
            invertIdMap.set(src, dst);
        }
        let changeEmptiesSelf = false;
        let changeFillsSelf = false;
        const invertedMoves = [];
        for (const [src, dst] of moves) {
            changeEmptiesSelf || (changeEmptiesSelf = src === "self");
            changeFillsSelf || (changeFillsSelf = dst === "self");
            /* eslint-disable tsdoc/syntax */
            /**
             * TODO:AB#6319: The targeting choices here are not really semantically right; this can lead to a situation where we put a node
             * in a non-empty register.
             *
             * Consider:
             * Start: field contents are "A". An edit E1 replaces "A" with "B". Then two branches with that edit on their undo-redo stack
             * concurrently undo it.
             *
             * Giving names to these edits:
             * E1: Replace A with B ( build [E1,1], move: cell:self->[E1,2], node:[E1,1]->self )
             * E2: undo E1 ( move: cell:self->[E1,1], node:[E1,2]->self )
             * E3, concurrent to E2: undo E1 ( move: cell:self->[E1,1], node:[E1,2]->self )
             *
             * Say E2 is sequenced first. Output context of E2 has registers "self" and [E1,1] filled with "A" and "B" respectively.
             *
             * Now consider E3' = rebase(E3 over E2).
             *
             * When rebasing the moves,
             * cell:self->[E1,1] remains the same as it targets the cell.
             * node:[E1,2]->self becomes self->self.
             *
             * This is an invalid edit for two reasons:
             * 1. It has two destinations for the "self" register
             * 2. One of those destinations is non-empty, without a corresponding move to empty it.
             */
            /* eslint-enable tsdoc/syntax */
            const target = src !== "self" && dst === "self" ? "cellTargeting" : "nodeTargeting";
            invertedMoves.push([withIntention(dst), withIntention(src), target]);
        }
        const inverted = {
            moves: invertedMoves,
            childChanges: childChanges.map(([id, childChange]) => [
                withIntention(invertIdMap.get(id) ?? id),
                invertChild(childChange),
            ]),
        };
        const inverseEmptiesSelf = changeFillsSelf;
        const inverseFillsSelf = changeEmptiesSelf;
        if (inverseFillsSelf && !inverseEmptiesSelf) {
            inverted.reservedDetachId = { localId: genId.getNextId() };
        }
        return inverted;
    },
    rebase: (change, overTagged, rebaseChild) => {
        const withIntention = (id) => {
            if (id === "self") {
                return id;
            }
            const intention = id.revision ?? overTagged.revision;
            return { revision: intention, localId: id.localId };
        };
        const { moves, childChanges } = change;
        const { change: overChange } = overTagged;
        const rebasedMoves = [];
        const overDstToSrc = new RegisterMap();
        const overSrcToDst = new RegisterMap();
        for (const [src, dst] of overChange.moves) {
            const srcTagged = withIntention(src);
            const dstTagged = withIntention(dst);
            overSrcToDst.set(srcTagged, dstTagged);
            overDstToSrc.set(dstTagged, srcTagged);
        }
        let reservedDetachId = change.reservedDetachId;
        const changeDstToSrc = new RegisterMap();
        const changeSrcToDst = new RegisterMap();
        for (const [src, dst] of moves) {
            changeSrcToDst.set(src, dst);
            changeDstToSrc.set(dst, src);
        }
        for (const [src, dst, target] of moves) {
            if (target === "cellTargeting") {
                (0, core_utils_1.assert)(src === "self", 0x857 /* Cell targeting moves must have self as a source */);
                // TODO: Should we drop cell targeting / node targeting and just special-case 'self'? Might be simpler to understand.
                // Holding off on making a call until AB#6298 is addressed (possibly support for rebasing transactions makes the
                // answer to this more obvious).
                if (overSrcToDst.get(src) !== undefined &&
                    overDstToSrc.get(src) === undefined &&
                    reservedDetachId === undefined) {
                    // Over removed the content occupying this cell and didn't fill it with other content.
                    // This change includes field changes,
                    reservedDetachId = dst;
                }
                else {
                    // Cell-targeting.
                    rebasedMoves.push([src, dst, target]);
                }
            }
            else {
                // Figure out where content in src ended up in `overTagged`
                const rebasedSrc = overSrcToDst.get(src) ?? src;
                // Note: we cannot drop changes which map a node to itself, as this loses the intention of the original edit
                // (since the target kind is node targeting, it may not still map to a noop after further rebases)
                rebasedMoves.push([rebasedSrc, dst, target]);
            }
        }
        // We rebased a fill from an empty state over another edit which also sets this field.
        // We need to make sure that we also empty the field.
        const overFillsEmptyField = !overSrcToDst.has("self") && overDstToSrc.has("self");
        if (overFillsEmptyField && reservedDetachId !== undefined) {
            rebasedMoves.push(["self", reservedDetachId, "cellTargeting"]);
            reservedDetachId = undefined;
        }
        const overChildChangesBySrc = new RegisterMap();
        for (const [id, childChange] of overChange.childChanges ?? []) {
            overChildChangesBySrc.set(withIntention(id) ?? id, childChange);
        }
        const rebasedChildChanges = [];
        for (const [id, childChange] of childChanges) {
            const overChildChange = overChildChangesBySrc.get(id);
            const rebasedId = overSrcToDst.get(id) ?? id;
            const rebasedChildChange = rebaseChild(childChange, overChildChange, rebasedId === "self" ? index_js_2.NodeExistenceState.Alive : index_js_2.NodeExistenceState.Dead);
            if (rebasedChildChange !== undefined) {
                rebasedChildChanges.push([rebasedId, rebasedChildChange]);
            }
        }
        const rebased = {
            moves: rebasedMoves,
            childChanges: rebasedChildChanges,
        };
        if (reservedDetachId !== undefined) {
            rebased.reservedDetachId = reservedDetachId;
        }
        return rebased;
    },
    prune: (change, pruneChild) => {
        const childChanges = [];
        const prunedChange = {
            moves: change.moves,
            childChanges,
        };
        if (change.reservedDetachId !== undefined) {
            prunedChange.reservedDetachId = change.reservedDetachId;
        }
        for (const [id, childChange] of change.childChanges) {
            const prunedChildChange = pruneChild(childChange);
            if (prunedChildChange !== undefined) {
                prunedChange.childChanges.push([id, prunedChildChange]);
            }
        }
        return prunedChange;
    },
};
function getBidirectionalMaps(moves, revision) {
    const srcToDst = new RegisterMap();
    const dstToSrc = new RegisterMap();
    for (const [src, dst, target] of moves) {
        const srcWithRevision = withRevision(src, revision);
        const dstWithRevision = withRevision(dst, revision);
        srcToDst.set(srcWithRevision, [dstWithRevision, target]);
        dstToSrc.set(dstWithRevision, srcWithRevision);
    }
    return { srcToDst, dstToSrc };
}
function withRevisionOrUndefined(id, revision) {
    return id !== undefined ? withRevision(id, revision) : undefined;
}
function withRevision(id, revision) {
    if (id === "self") {
        return id;
    }
    return { revision: id.revision ?? revision, localId: id.localId };
}
exports.optionalFieldEditor = {
    set: (wasEmpty, ids) => {
        const result = {
            moves: [[{ localId: ids.fill }, "self", "nodeTargeting"]],
            childChanges: [],
        };
        if (wasEmpty) {
            result.reservedDetachId = { localId: ids.detach };
        }
        else {
            result.moves.push(["self", { localId: ids.detach }, "cellTargeting"]);
        }
        return result;
    },
    clear: (wasEmpty, detachId) => wasEmpty
        ? { moves: [], childChanges: [], reservedDetachId: { localId: detachId } }
        : {
            moves: [["self", { localId: detachId }, "cellTargeting"]],
            childChanges: [],
        },
    buildChildChange: (index, childChange) => {
        (0, core_utils_1.assert)(index === 0, 0x404 /* Optional fields only support a single child node */);
        return {
            moves: [],
            childChanges: [["self", childChange]],
        };
    },
};
function optionalFieldIntoDelta({ change, revision }, deltaFromChild) {
    const delta = {};
    let markIsANoop = true;
    const mark = { count: 1 };
    if (change.moves.length > 0) {
        const renames = [];
        for (const [src, dst] of change.moves) {
            if (src === "self" && dst !== "self") {
                mark.detach = { major: dst.revision ?? revision, minor: dst.localId };
                markIsANoop = false;
            }
            else if (dst === "self" && src !== "self") {
                mark.attach = { major: src.revision ?? revision, minor: src.localId };
                markIsANoop = false;
            }
            else if (src !== "self" && dst !== "self") {
                renames.push({
                    count: 1,
                    oldId: { major: src.revision ?? revision, minor: src.localId },
                    newId: { major: dst.revision ?? revision, minor: dst.localId },
                });
            }
        }
        if (renames.length > 0) {
            delta.rename = renames;
        }
    }
    if (change.childChanges.length > 0) {
        const globals = [];
        for (const [id, childChange] of change.childChanges) {
            const childDelta = deltaFromChild(childChange);
            if (id !== "self") {
                const fields = childDelta;
                globals.push({
                    id: { major: id.revision ?? revision, minor: id.localId },
                    fields,
                });
            }
            else {
                mark.fields = childDelta;
                markIsANoop = false;
            }
        }
        if (globals.length > 0) {
            delta.global = globals;
        }
    }
    if (!markIsANoop) {
        delta.local = [mark];
    }
    return delta;
}
exports.optionalFieldIntoDelta = optionalFieldIntoDelta;
exports.optionalChangeHandler = {
    rebaser: exports.optionalChangeRebaser,
    codecsFactory: optionalFieldCodecs_js_1.makeOptionalFieldCodecFamily,
    editor: exports.optionalFieldEditor,
    intoDelta: optionalFieldIntoDelta,
    relevantRemovedRoots,
    isEmpty: (change) => change.childChanges.length === 0 &&
        change.moves.length === 0 &&
        change.reservedDetachId === undefined,
    createEmpty: () => ({ moves: [], childChanges: [] }),
};
function* relevantRemovedRoots({ change, revision }, relevantRemovedRootsFromChild) {
    const alreadyYielded = new RegisterMap();
    for (const [src] of change.moves) {
        if (src !== "self" && !alreadyYielded.has(src)) {
            alreadyYielded.set(src, true);
            yield (0, deltaUtils_js_1.nodeIdFromChangeAtom)(src, revision);
        }
    }
    for (const [id, childChange] of change.childChanges) {
        // Child changes make the tree they apply to relevant unless that tree existed in the starting context of
        // of this change.
        if (id !== "self" && !alreadyYielded.has(id)) {
            alreadyYielded.set(id, true);
            yield (0, deltaUtils_js_1.nodeIdFromChangeAtom)(id);
        }
        yield* relevantRemovedRootsFromChild(childChange);
    }
}
//# sourceMappingURL=optionalField.js.map