"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.SharedTreeFactory = exports.defaultSharedTreeOptions = exports.ForestType = exports.SharedTree = void 0;
const core_utils_1 = require("@fluidframework/core-utils");
const telemetry_utils_1 = require("@fluidframework/telemetry-utils");
const index_js_1 = require("../codec/index.js");
const index_js_2 = require("../core/index.js");
const index_js_3 = require("../shared-tree-core/index.js");
const index_js_4 = require("../feature-libraries/index.js");
const index_js_5 = require("../events/index.js");
const index_js_6 = require("../util/index.js");
const index_js_7 = require("../simple-tree/index.js");
const schematizedTree_js_1 = require("./schematizedTree.js");
const treeCheckout_js_1 = require("./treeCheckout.js");
const treeView_js_1 = require("./treeView.js");
const sharedTreeChangeFamily_js_1 = require("./sharedTreeChangeFamily.js");
/**
 * Shared tree, configured with a good set of indexes and field kinds which will maintain compatibility over time.
 *
 * TODO: detail compatibility requirements.
 */
class SharedTree extends index_js_3.SharedTreeCore {
    get storedSchema() {
        return this.checkout.storedSchema;
    }
    constructor(id, runtime, attributes, optionsParam, telemetryContextPrefix) {
        (0, core_utils_1.assert)(runtime.idCompressor !== undefined, 0x883 /* IdCompressor must be enabled to use SharedTree */);
        const options = { ...exports.defaultSharedTreeOptions, ...optionsParam };
        const schema = new index_js_2.TreeStoredSchemaRepository();
        const forest = options.forest === ForestType.Optimized
            ? (0, index_js_4.buildChunkedForest)((0, index_js_4.makeTreeChunker)(schema, index_js_4.defaultSchemaPolicy))
            : (0, index_js_4.buildForest)();
        const revisionTagCodec = new index_js_2.RevisionTagCodec(runtime.idCompressor);
        const removedRoots = (0, index_js_2.makeDetachedFieldIndex)("repair", revisionTagCodec, options);
        const schemaSummarizer = new index_js_4.SchemaSummarizer(runtime, schema, options, {
            getCurrentSeq: () => this.runtime.deltaManager.lastSequenceNumber,
        });
        const fieldBatchCodec = (0, index_js_4.makeFieldBatchCodec)(options);
        const encoderContext = {
            schema: {
                schema,
                policy: index_js_4.defaultSchemaPolicy,
            },
            encodeType: options.treeEncodeType,
        };
        const forestSummarizer = new index_js_4.ForestSummarizer(forest, revisionTagCodec, fieldBatchCodec, encoderContext, options);
        const removedRootsSummarizer = new index_js_4.DetachedFieldIndexSummarizer(removedRoots);
        const innerChangeFamily = new sharedTreeChangeFamily_js_1.SharedTreeChangeFamily(revisionTagCodec, fieldBatchCodec, options);
        const changeFamily = (0, index_js_4.makeMitigatedChangeFamily)(innerChangeFamily, sharedTreeChangeFamily_js_1.SharedTreeChangeFamily.emptyChange, (error) => {
            // TODO:6344 Add telemetry for these errors.
            // Rethrowing the error has a different effect depending on the context in which the
            // ChangeFamily was invoked:
            // - If the ChangeFamily was invoked as part of incoming op processing, rethrowing the error
            // will cause the runtime to disconnect the client, log a severe error, and not reconnect.
            // This will not cause the host application to crash because it is not on the stack at that time.
            // TODO: let the host application know that the client is now disconnected.
            // - If the ChangeFamily was invoked as part of dealing with a local change, rethrowing the
            // error will cause the host application to crash. This is not ideal, but is better than
            // letting the application either send an invalid change to the server or allowing the
            // application to continue working when its local branches contain edits that cannot be
            // reflected in its views.
            // The best course of action for a host application in such a state is to restart.
            // TODO: let the host application know about this situation and provide a way to
            // programmatically reload the SharedTree container.
            throw error;
        });
        super([schemaSummarizer, forestSummarizer, removedRootsSummarizer], changeFamily, options, id, runtime, attributes, telemetryContextPrefix);
        /**
         * Creating multiple editable tree contexts for the same branch, and thus with the same underlying AnchorSet does not work due to how TreeNode caching works.
         * This flag is used to detect if one already exists for the main branch and error if creating a second.
         * THis should catch most accidental violations of this restriction but there are still ways to create two conflicting contexts (for example calling constructing one manually).
         *
         * TODO:
         * 1. API docs need to reflect this limitation or the limitation has to be removed.
         */
        this.hasView = false;
        this._events = (0, index_js_5.createEmitter)();
        const localBranch = this.getLocalBranch();
        this.checkout = (0, treeCheckout_js_1.createTreeCheckout)(runtime.idCompressor, revisionTagCodec, {
            branch: localBranch,
            changeFamily,
            schema,
            forest,
            fieldBatchCodec,
            events: this._events,
            removedRoots,
        });
    }
    requireSchema(schema, onSchemaIncompatible, nodeKeyManager, nodeKeyFieldKey) {
        (0, core_utils_1.assert)(this.hasView === false, 0x7f1 /* Cannot create second view from tree. */);
        const viewSchema = new index_js_4.ViewSchema(index_js_4.defaultSchemaPolicy, {}, schema);
        const compatibility = viewSchema.checkCompatibility(this.storedSchema);
        if (compatibility.write !== index_js_2.Compatibility.Compatible ||
            compatibility.read !== index_js_2.Compatibility.Compatible) {
            return undefined;
        }
        this.hasView = true;
        const view = new treeView_js_1.CheckoutFlexTreeView(this.checkout, schema, nodeKeyManager ?? (0, index_js_4.createNodeKeyManager)(this.runtime.idCompressor), nodeKeyFieldKey ?? (0, index_js_6.brand)(index_js_4.nodeKeyFieldKey), () => {
            (0, core_utils_1.assert)(this.hasView, 0x7f2 /* unexpected dispose */);
            this.hasView = false;
        });
        const onSchemaChange = () => {
            const compatibilityInner = viewSchema.checkCompatibility(this.storedSchema);
            if (compatibilityInner.write !== index_js_2.Compatibility.Compatible ||
                compatibilityInner.read !== index_js_2.Compatibility.Compatible) {
                view[index_js_6.disposeSymbol]();
                onSchemaIncompatible();
                return false;
            }
            else {
                return true;
            }
        };
        (0, schematizedTree_js_1.afterSchemaChanges)(this._events, this.checkout, onSchemaChange);
        return view;
    }
    contentSnapshot() {
        const cursor = this.checkout.forest.allocateCursor();
        try {
            (0, index_js_2.moveToDetachedField)(this.checkout.forest, cursor);
            return {
                schema: this.storedSchema.clone(),
                tree: (0, index_js_4.jsonableTreeFromFieldCursor)(cursor),
                removed: this.checkout.getRemovedRoots(),
            };
        }
        finally {
            cursor.free();
        }
    }
    schematizeInternal(config, nodeKeyManager, nodeKeyFieldKey) {
        if (this.hasView === true) {
            throw new telemetry_utils_1.UsageError("Only one view can be constructed from a given tree at a time. Dispose of the first before creating a second.");
        }
        // TODO:
        // When this becomes a more proper out of schema adapter, editing should be made lazy.
        // This will improve support for readonly documents, cross version collaboration and attribution.
        // Check for empty.
        if (this.checkout.forest.isEmpty && (0, index_js_2.schemaDataIsEmpty)(this.storedSchema)) {
            this.checkout.transaction.start();
            (0, schematizedTree_js_1.initializeContent)(this.checkout, config.schema, () => {
                const field = { field: index_js_2.rootFieldKey, parent: undefined };
                const content = (0, index_js_4.normalizeNewFieldContent)({ schema: config.schema }, config.schema.rootFieldSchema, config.initialTree);
                switch (this.storedSchema.rootFieldSchema.kind.identifier) {
                    case index_js_4.FieldKinds.optional.identifier: {
                        const fieldEditor = this.editor.optionalField(field);
                        (0, core_utils_1.assert)(content.getFieldLength() <= 1, 0x7f4 /* optional field content should normalize at most one item */);
                        fieldEditor.set(content.getFieldLength() === 0 ? undefined : content, true);
                        break;
                    }
                    case index_js_4.FieldKinds.sequence.identifier: {
                        const fieldEditor = this.editor.sequenceField(field);
                        // TODO: should do an idempotent edit here.
                        fieldEditor.insert(0, content);
                        break;
                    }
                    default: {
                        (0, index_js_6.fail)("unexpected root field kind during initialize");
                    }
                }
            });
            this.checkout.transaction.commit();
        }
        (0, schematizedTree_js_1.schematize)(this.checkout.events, this.checkout, config);
        return (this.requireSchema(config.schema, () => (0, index_js_6.fail)("schema incompatible"), nodeKeyManager, nodeKeyFieldKey) ?? (0, index_js_6.fail)("Schematize failed"));
    }
    schematize(config) {
        const flexConfig = (0, index_js_7.toFlexConfig)(config);
        const view = this.schematizeInternal(flexConfig);
        return new index_js_7.WrapperTreeView(view);
    }
    async loadCore(services) {
        await super.loadCore(services);
        this._events.emit("afterBatch");
    }
}
exports.SharedTree = SharedTree;
/**
 * Used to distinguish between different forest types.
 * @internal
 */
var ForestType;
(function (ForestType) {
    /**
     * The "ObjectForest" forest type.
     */
    ForestType[ForestType["Reference"] = 0] = "Reference";
    /**
     * The "ChunkedForest" forest type.
     */
    ForestType[ForestType["Optimized"] = 1] = "Optimized";
})(ForestType || (exports.ForestType = ForestType = {}));
exports.defaultSharedTreeOptions = {
    jsonValidator: index_js_1.noopValidator,
    forest: ForestType.Reference,
    treeEncodeType: index_js_4.TreeCompressionStrategy.Compressed,
};
/**
 * A channel factory that creates {@link ISharedTree}s.
 * @internal
 */
class SharedTreeFactory {
    constructor(options = {}) {
        this.options = options;
        this.type = "https://graph.microsoft.com/types/tree";
        this.attributes = {
            type: this.type,
            snapshotFormatVersion: "0.0.0",
            packageVersion: "0.0.0",
        };
    }
    async load(runtime, id, services, channelAttributes) {
        const tree = new SharedTree(id, runtime, channelAttributes, this.options, "SharedTree");
        await tree.load(services);
        return tree;
    }
    create(runtime, id) {
        const tree = new SharedTree(id, runtime, this.attributes, this.options, "SharedTree");
        tree.initializeLocal();
        return tree;
    }
}
exports.SharedTreeFactory = SharedTreeFactory;
//# sourceMappingURL=sharedTree.js.map