"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.structuralName = exports.createTree = exports.SchemaFactory = exports.schemaFromValue = void 0;
const core_utils_1 = require("@fluidframework/core-utils");
const index_js_1 = require("../util/index.js");
const index_js_2 = require("../feature-libraries/index.js");
const index_js_3 = require("../domains/index.js");
const proxies_js_1 = require("./proxies.js");
const toFlexSchema_js_1 = require("./toFlexSchema.js");
const schemaTypes_js_1 = require("./schemaTypes.js");
const types_js_1 = require("./types.js");
/**
 * Instances of this class are schema for leaf nodes.
 * @remarks
 * Unlike other schema, leaf schema are class instances instead of classes themselves.
 * This is because the instance type (the tree node type) for leaves are not objects,
 * so those instances can't be instances of a schema based class.
 * @privateRemarks
 * This class refers to the underlying flex tree schema in its constructor, so this class can't be included in the package API.
 */
class LeafNodeSchema {
    create(data) {
        return data;
    }
    constructor(schema) {
        this.kind = schemaTypes_js_1.NodeKind.Leaf;
        this.implicitlyConstructable = true;
        (0, toFlexSchema_js_1.setFlexSchemaFromClassSchema)(this, schema);
        this.identifier = schema.name;
        this.info = schema.info;
    }
}
/**
 * Wrapper around LeafNodeSchema's constructor that provides the return type that is desired in the package public API.
 */
function makeLeaf(schema) {
    return new LeafNodeSchema(schema);
}
// Leaf schema shared between all SchemaFactory instances.
const stringSchema = makeLeaf(index_js_3.leaf.string);
const numberSchema = makeLeaf(index_js_3.leaf.number);
const booleanSchema = makeLeaf(index_js_3.leaf.boolean);
const nullSchema = makeLeaf(index_js_3.leaf.null);
const handleSchema = makeLeaf(index_js_3.leaf.handle);
/**
 * Gets the leaf domain schema compatible with a given {@link TreeValue}.
 */
function schemaFromValue(value) {
    switch (typeof value) {
        case "boolean":
            return booleanSchema;
        case "number":
            return numberSchema;
        case "string":
            return stringSchema;
        case "object": {
            if (value === null) {
                return nullSchema;
            }
            (0, core_utils_1.assert)((0, index_js_2.isFluidHandle)(value), 0x87e /* invalid TreeValue */);
            return handleSchema;
        }
        default:
            (0, core_utils_1.unreachableCase)(value);
    }
}
exports.schemaFromValue = schemaFromValue;
/**
 * Builds schema libraries, and the schema within them.
 *
 * @typeParam TScope - Scope added as a prefix to the name of every schema produced by this factory.
 * @typeParam TName - Type of names used to identify each schema produced in this factory.
 * Typically this is just `string` but it is also possible to use `string` or `number` based enums if you prefer to identify your types that way.
 *
 * @sealed @public
 */
class SchemaFactory {
    /**
     * @param scope - Prefix appended to the identifiers of all {@link TreeNodeSchema} produced by this builder.
     * Use of [Reverse domain name notation](https://en.wikipedia.org/wiki/Reverse_domain_name_notation) or a UUIDv4 is recommended to avoid collisions.
     */
    constructor(scope) {
        this.scope = scope;
        this.structuralTypes = new Map();
        /**
         * {@link TreeNodeSchema} for holding a JavaScript `string`.
         *
         * @remarks
         * Strings containing unpaired UTF-16 surrogate pair code units may not be handled correctly.
         *
         * These limitations come from the use of UTF-8 encoding of the strings, which requires them to be valid unicode.
         * JavaScript does not make this requirement for its strings so not all possible JavaScript strings are supported.
         * @privateRemarks
         * TODO:
         * We should be much more clear about what happens if you use problematic values.
         * We should validate and/or normalize them when inserting content.
         */
        this.string = stringSchema;
        /**
         * {@link TreeNodeSchema} for holding a JavaScript `number`.
         *
         * @remarks
         * The number is a [double-precision 64-bit binary format IEEE 754](https://en.wikipedia.org/wiki/Double-precision_floating-point_format) value, however there are some exceptions:
         * - `NaN`, and the infinities are converted to `null` (and may therefore only be used where `null` is allowed by the schema).
         * - `-0` may be converted to `0` in some cases.
         *
         * These limitations match the limitations of JSON.
         * @privateRemarks
         * TODO:
         * We should be much more clear about what happens if you use problematic values.
         * We should validate and/or normalize them when inserting content.
         */
        this.number = numberSchema;
        /**
         * {@link TreeNodeSchema} for holding a boolean.
         */
        this.boolean = booleanSchema;
        /**
         * {@link TreeNodeSchema} for JavaScript `null`.
         *
         * @remarks
         * There are good [reasons to avoid using null](https://www.npmjs.com/package/%40rushstack/eslint-plugin#rushstackno-new-null) in JavaScript, however sometimes it is desired.
         * This {@link TreeNodeSchema} node provides the option to include nulls in trees when desired.
         * Unless directly inter-operating with existing data using null, consider other approaches, like wrapping the value in an optional field, or using a more specifically named empty object node.
         */
        this.null = nullSchema;
        /**
         * {@link TreeNodeSchema} for holding an {@link @fluidframework/core-interfaces#(IFluidHandle:interface)}.
         */
        this.handle = handleSchema;
    }
    scoped(name) {
        return `${this.scope}.${name}`;
    }
    /**
     * Construct a class that provides the common parts all TreeNodeSchemaClass share.
     * More specific schema extend this class.
     */
    nodeSchema(name, kind, t, implicitlyConstructable) {
        const identifier = this.scoped(name);
        class schema extends types_js_1.TreeNode {
            /**
             * This constructor only does validation of the input, and should be passed the argument from the derived type unchanged.
             * It is up to the derived type to actually do something with this value.
             */
            constructor(input) {
                super();
                // Currently this just does validation. All other logic is in the subclass.
                if ((0, index_js_2.isFlexTreeNode)(input)) {
                    (0, core_utils_1.assert)((0, proxies_js_1.getClassSchema)(input.schema) === this.constructor, 0x83b /* building node with wrong schema */);
                }
                // TODO: make this a better user facing error, and explain how to copy explicitly.
                (0, core_utils_1.assert)(!(input instanceof types_js_1.TreeNode), 0x83c /* Existing nodes cannot be used as new content to insert. They must either be moved or explicitly copied */);
            }
            get [schemaTypes_js_1.type]() {
                return identifier;
            }
        }
        schema.identifier = identifier;
        schema.kind = kind;
        schema.info = t;
        schema.implicitlyConstructable = implicitlyConstructable;
        // Class objects are functions (callable), so we need a strong way to distinguish between `schema` and `() => schema` when used as a `LazyItem`.
        (0, index_js_2.markEager)(schema);
        return schema;
    }
    /**
     * Define a {@link TreeNodeSchema} for an object node.
     *
     * @param name - Unique identifier for this schema within this factory's scope.
     */
    object(name, t) {
        const allowAdditionalProperties = true;
        class schema extends this.nodeSchema(name, schemaTypes_js_1.NodeKind.Object, t, true) {
            constructor(input) {
                super(input);
                if ((0, index_js_2.isFlexTreeNode)(input)) {
                    return (0, proxies_js_1.createNodeProxy)(input, allowAdditionalProperties, this);
                }
                else {
                    const flexSchema = (0, toFlexSchema_js_1.getFlexSchema)(this.constructor);
                    return (0, proxies_js_1.createRawNodeProxy)(flexSchema, input, allowAdditionalProperties, this);
                }
            }
        }
        return schema;
    }
    map(nameOrAllowedTypes, allowedTypes) {
        if (allowedTypes === undefined) {
            const types = nameOrAllowedTypes;
            const fullName = structuralName("Map", types);
            return (0, index_js_1.getOrCreate)(this.structuralTypes, fullName, () => this.namedMap_internal(fullName, nameOrAllowedTypes, false, true));
        }
        return this.namedMap_internal(nameOrAllowedTypes, allowedTypes, true, true);
    }
    /**
     * Define a {@link TreeNodeSchema} for a {@link (TreeMapNode:interface)}.
     *
     * @param name - Unique identifier for this schema within this factory's scope.
     *
     * @remarks See remarks on {@link SchemaFactory.namedArray_internal}.
     */
    namedMap_internal(name, allowedTypes, customizable, implicitlyConstructable) {
        class schema extends this.nodeSchema(name, schemaTypes_js_1.NodeKind.Map, allowedTypes, implicitlyConstructable) {
            constructor(input) {
                super(input);
                if ((0, index_js_2.isFlexTreeNode)(input)) {
                    return (0, proxies_js_1.createNodeProxy)(input, customizable, customizable ? this : undefined);
                }
                else {
                    const flexSchema = (0, toFlexSchema_js_1.getFlexSchema)(this.constructor);
                    return (0, proxies_js_1.createRawNodeProxy)(flexSchema, input, customizable, customizable ? this : undefined);
                }
            }
        }
        // Setup map functionality
        Object.defineProperties(schema.prototype, proxies_js_1.mapStaticDispatchMap);
        return schema;
    }
    array(nameOrAllowedTypes, allowedTypes) {
        if (allowedTypes === undefined) {
            const types = nameOrAllowedTypes;
            const fullName = structuralName("Array", types);
            return (0, index_js_1.getOrCreate)(this.structuralTypes, fullName, () => this.namedArray_internal(fullName, nameOrAllowedTypes, false, true));
        }
        return this.namedArray_internal(nameOrAllowedTypes, allowedTypes, true, true);
    }
    /**
     * Define a {@link TreeNodeSchema} for a {@link (TreeArrayNode:interface)}.
     *
     * @param name - Unique identifier for this schema within this factory's scope.
     *
     * @remarks
     * This is not intended to be used directly, use the overload of `array` which takes a name instead.
     * This is only public to work around a compiler limitation.
     *
     * @privateRemarks
     * TODO: this should be made private or protected.
     * Doing so breaks due to:
     * `src/class-tree/schemaFactoryRecursive.ts:42:9 - error TS2310: Type 'Array' recursively references itself as a base type.`
     * Once recursive APIs are better sorted out and integrated into this class, switch this back to private.
     */
    namedArray_internal(name, allowedTypes, customizable, implicitlyConstructable) {
        // This class returns a proxy from its constructor to handle numeric indexing.
        // Alternatively it could extend a normal class which gets tons of numeric properties added.
        class schema extends this.nodeSchema(name, schemaTypes_js_1.NodeKind.Array, allowedTypes, implicitlyConstructable) {
            get length() {
                return (0, proxies_js_1.getSequenceField)(this).length;
            }
            constructor(input) {
                super(input);
                if ((0, index_js_2.isFlexTreeNode)(input)) {
                    return (0, proxies_js_1.createNodeProxy)(input, customizable, customizable ? this : undefined);
                }
                else {
                    const flexSchema = (0, toFlexSchema_js_1.getFlexSchema)(this.constructor);
                    return (0, proxies_js_1.createRawNodeProxy)(flexSchema, [...input], customizable, customizable ? this : undefined);
                }
            }
        }
        // Setup array functionality
        Object.defineProperties(schema.prototype, proxies_js_1.arrayNodePrototypeProperties);
        return schema;
    }
    /**
     * Make a field optional instead of the default which is required.
     */
    optional(t) {
        return new schemaTypes_js_1.FieldSchema(schemaTypes_js_1.FieldKind.Optional, t);
    }
    /**
     * Function which can be used for its compile time side-effects to tweak the evaluation order of recursive types to make them compile.
     * @remarks
     * Some related information in https://github.com/microsoft/TypeScript/issues/55758.
     *
     * Also be aware that code which relies on this tends to break VSCode's IntelliSense every time anything related to that code (even comments) is edited.
     * Running the command `TypeScript: Restart TS Server` with the schema file focused should fix it.
     * Sometimes this does not work: closing all open files except the schema before running the command can help.
     * Real compile errors (for example elsewhere in the file) can also cause the IntelliSense to not work correctly ever after `TypeScript: Restart TS Server`.
     *
     * Intellisense has also shown problems when schema files with recursive types are part of a cyclic file dependency.
     * Splitting the schema into its own file with minimal dependencies can help with this.
     *
     * Ensure `"noImplicitAny": true` is set in the `tsconfig.json`.
     * Without it, recursive types that are not working properly can infer `any` and give very non-type-safe results instead of erroring.
     *
     * @example
     * ```typescript
     * const factory = new SchemaFactory("example");
     * const recursiveReference = () => RecursiveObject;
     * factory.fixRecursiveReference(recursiveReference);
     * export class RecursiveObject extends factory.object("exampleObject", {
     * 	recursive: [recursiveReference],
     * }) {}
     * ```
     */
    fixRecursiveReference(...types) { }
}
exports.SchemaFactory = SchemaFactory;
// TODO: unify this with logic in getOrCreateNodeProxy
function createTree(schema, data) {
    if (typeof schema === "function") {
        return new schema(data);
    }
    return schema.create(data);
}
exports.createTree = createTree;
function structuralName(collectionName, allowedTypes) {
    let inner;
    if (!(0, index_js_1.isReadonlyArray)(allowedTypes)) {
        return structuralName(collectionName, [allowedTypes]);
    }
    else {
        const names = allowedTypes.map((t) => {
            // Ensure that lazy types (functions) don't slip through here.
            (0, core_utils_1.assert)(!(0, index_js_2.isLazy)(t), 0x83d /* invalid type provided */);
            return t.identifier;
        });
        // Ensure name is order independent
        names.sort();
        // Ensure name can't have collisions by quoting and escaping any quotes in the names of types.
        // Using JSON is a simple way to accomplish this.
        // The outer `[]` around the result are also needed so that a single type name "Any" would not collide with the "any" case above.
        inner = JSON.stringify(names);
    }
    return `${collectionName}<${inner}>`;
}
exports.structuralName = structuralName;
//# sourceMappingURL=schemaFactory.js.map